"""Project and Dashboard API routes for Knowledge Repository v3."""

from fastapi import APIRouter, HTTPException
from typing import List, Optional

from ..models import (
    ProjectDataModel,
    ProjectSummaryModel,
    ProjectListResponse,
    DashboardStatsModel,
    ActivityModel,
    ActivityFeedResponse,
    PersonalNoteModel,
    AddNoteRequest,
    UpdateNoteRequest,
    UpdateTagsRequest,
    KnowledgeGraphModel,
    TopicModel,
    TopicsResponse,
    JournalEntryModel,
    JournalResponse,
    ExtractRequest,
)
from ..services.project import (
    get_project,
    save_project,
    delete_project,
    load_projects,
    add_note_to_project,
    update_note,
    delete_note,
    toggle_star,
    update_tags,
    get_dashboard_stats,
    get_recent_activity,
    ProjectData,
)
from ..services.knowledge_graph import (
    load_graph,
    build_graph_from_projects,
    get_graph_stats,
    get_topic_clusters,
    find_related_videos,
    search_graph,
)
from ..services.history import (
    migrate_history_to_projects,
    get_or_create_project,
)
from collections import defaultdict

router = APIRouter()


# =============================================================================
# Dashboard Endpoints
# =============================================================================

@router.get("/api/dashboard/stats", response_model=DashboardStatsModel)
async def get_stats():
    """Get dashboard statistics."""
    stats = get_dashboard_stats()
    return DashboardStatsModel(**stats)


@router.get("/api/dashboard/activity", response_model=ActivityFeedResponse)
async def get_activity(limit: int = 20):
    """Get recent activity feed."""
    activities = get_recent_activity(limit)
    return ActivityFeedResponse(
        activities=[ActivityModel(**a) for a in activities]
    )


@router.post("/api/dashboard/migrate")
async def migrate_history():
    """Migrate v2 history items to v3 project format."""
    result = migrate_history_to_projects()
    return result


# =============================================================================
# Project List Endpoints
# =============================================================================

@router.get("/api/projects", response_model=ProjectListResponse)
async def list_projects(
    search: Optional[str] = None,
    starred: Optional[bool] = None,
    tag: Optional[str] = None,
    limit: int = 50,
    offset: int = 0
):
    """List all projects with optional filtering."""
    projects = load_projects()

    # Apply filters
    if search:
        search_lower = search.lower()
        projects = [
            p for p in projects
            if search_lower in p.get('title', '').lower()
            or search_lower in p.get('channel', '').lower()
            or any(search_lower in t.lower() for t in p.get('tags', []))
        ]

    if starred is not None:
        projects = [p for p in projects if p.get('is_starred') == starred]

    if tag:
        projects = [p for p in projects if tag in p.get('tags', [])]

    total = len(projects)
    projects = projects[offset:offset + limit]

    # Convert to summary models
    summaries = []
    for p in projects:
        project_data = ProjectData.from_dict(p)
        summaries.append(ProjectSummaryModel(
            video_id=project_data.video_id,
            title=project_data.title,
            url=project_data.url,
            channel=project_data.channel,
            duration=project_data.duration,
            thumbnail=project_data.thumbnail,
            is_starred=project_data.is_starred,
            tags=project_data.tags,
            created_at=project_data.created_at,
            extraction_status=project_data.extraction_status,
            insight_count=len(project_data.insights),
            entity_count=project_data.entities.count(),
            note_count=len(project_data.notes)
        ))

    return ProjectListResponse(projects=summaries, total=total)


# =============================================================================
# Single Project Endpoints
# =============================================================================

@router.get("/api/project/{video_id}")
async def get_project_data(video_id: str):
    """Get complete project data."""
    # First try to get from projects, then fallback to history
    project = get_project(video_id)

    if not project:
        # Try to create from history
        project_dict = get_or_create_project(video_id)
        if not project_dict:
            raise HTTPException(status_code=404, detail="Project not found")
        return project_dict

    return project.to_dict()


@router.delete("/api/project/{video_id}")
async def delete_project_endpoint(video_id: str):
    """Delete a project."""
    success = delete_project(video_id)
    if not success:
        raise HTTPException(status_code=404, detail="Project not found")
    return {"success": True}


@router.post("/api/project/{video_id}/star")
async def toggle_star_endpoint(video_id: str):
    """Toggle star status for a project."""
    new_status = toggle_star(video_id)
    if new_status is None:
        raise HTTPException(status_code=404, detail="Project not found")
    return {"is_starred": new_status}


@router.put("/api/project/{video_id}/tags")
async def update_tags_endpoint(video_id: str, request: UpdateTagsRequest):
    """Update tags for a project."""
    result = update_tags(video_id, request.tags)
    if result is None:
        raise HTTPException(status_code=404, detail="Project not found")
    return {"tags": result}


# =============================================================================
# Personal Notes Endpoints
# =============================================================================

@router.get("/api/project/{video_id}/notes")
async def get_notes(video_id: str):
    """Get all notes for a project."""
    project = get_project(video_id)
    if not project:
        raise HTTPException(status_code=404, detail="Project not found")

    return {
        "notes": [n.to_dict() for n in project.notes]
    }


@router.post("/api/project/{video_id}/notes")
async def add_note(video_id: str, request: AddNoteRequest):
    """Add a note to a project."""
    note = add_note_to_project(
        video_id,
        request.content,
        request.timestamp,
        request.tags
    )
    if not note:
        raise HTTPException(status_code=404, detail="Project not found")

    return note.to_dict()


@router.put("/api/project/{video_id}/notes/{note_id}")
async def update_note_endpoint(video_id: str, note_id: str, request: UpdateNoteRequest):
    """Update a note."""
    note = update_note(video_id, note_id, request.content, request.tags)
    if not note:
        raise HTTPException(status_code=404, detail="Note not found")

    return note.to_dict()


@router.delete("/api/project/{video_id}/notes/{note_id}")
async def delete_note_endpoint(video_id: str, note_id: str):
    """Delete a note."""
    success = delete_note(video_id, note_id)
    if not success:
        raise HTTPException(status_code=404, detail="Note not found")

    return {"success": True}


# =============================================================================
# Connections Endpoints
# =============================================================================

@router.get("/api/project/{video_id}/connections")
async def get_connections(video_id: str, limit: int = 10):
    """Get related videos for a project."""
    connections = find_related_videos(video_id, limit)
    return {"connections": connections}


# =============================================================================
# Knowledge Graph Endpoints
# =============================================================================

@router.get("/api/graph")
async def get_knowledge_graph():
    """Get the complete knowledge graph."""
    graph = load_graph()
    return graph.to_dict()


@router.post("/api/graph/rebuild")
async def rebuild_graph():
    """Rebuild the knowledge graph from all projects."""
    graph = build_graph_from_projects()
    stats = get_graph_stats()
    return {
        "success": True,
        "stats": stats
    }


@router.get("/api/graph/stats")
async def get_graph_statistics():
    """Get knowledge graph statistics."""
    return get_graph_stats()


@router.get("/api/graph/search")
async def search_knowledge_graph(query: str, limit: int = 20):
    """Search the knowledge graph."""
    results = search_graph(query)[:limit]
    return {"results": results}


# =============================================================================
# Topics Endpoints
# =============================================================================

@router.get("/api/topics", response_model=TopicsResponse)
async def get_topics():
    """Get all topics with their video counts."""
    clusters = get_topic_clusters()

    topics = [
        TopicModel(
            name=c['topic'],
            video_count=c['video_count'],
            videos=[v['video_id'] for v in c['videos']]
        )
        for c in clusters
    ]

    return TopicsResponse(topics=topics, total=len(topics))


@router.get("/api/topics/{topic}")
async def get_topic_detail(topic: str):
    """Get details for a specific topic."""
    clusters = get_topic_clusters()

    for c in clusters:
        if c['topic'].lower() == topic.lower():
            return c

    raise HTTPException(status_code=404, detail="Topic not found")


# =============================================================================
# Learning Journal Endpoints
# =============================================================================

@router.get("/api/journal", response_model=JournalResponse)
async def get_journal(limit: int = 30):
    """Get learning journal entries grouped by date."""
    activities = get_recent_activity(limit * 5)  # Get more to group by date

    # Group by date
    by_date = defaultdict(list)
    for activity in activities:
        timestamp = activity.get('timestamp', '')
        if timestamp:
            date = timestamp.split('T')[0]  # Extract date part
            by_date[date].append(ActivityModel(**activity))

    # Convert to journal entries
    entries = [
        JournalEntryModel(date=date, entries=activities)
        for date, activities in sorted(by_date.items(), reverse=True)
    ][:limit]

    return JournalResponse(entries=entries)


# =============================================================================
# Extraction Pipeline Endpoints
# =============================================================================

@router.post("/api/project/{video_id}/extract")
async def run_extraction(video_id: str, request: ExtractRequest = None):
    """
    Run the extraction pipeline on a project.

    Extracts entities, insights, quotes, facts, and semantic sections.
    """
    from ..services.extraction import run_extraction_and_save

    # Get request parameters or use defaults
    if request is None:
        request = ExtractRequest()

    try:
        result = await run_extraction_and_save(
            video_id=video_id,
            extract_entities_flag=request.extract_entities,
            extract_insights_flag=request.extract_insights,
            extract_quotes_flag=request.extract_quotes,
            extract_facts_flag=request.extract_facts,
            extract_sections_flag=request.extract_sections
        )

        if "error" in result and result["error"] == "Project not found":
            raise HTTPException(status_code=404, detail="Project not found")

        return result

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@router.post("/api/project/{video_id}/extract/entities")
async def extract_entities_only(video_id: str):
    """Extract only entities from a project."""
    from ..services.extraction import run_extraction_and_save

    result = await run_extraction_and_save(
        video_id=video_id,
        extract_entities_flag=True,
        extract_insights_flag=False,
        extract_quotes_flag=False,
        extract_facts_flag=False,
        extract_sections_flag=False
    )

    if "error" in result and result["error"] == "Project not found":
        raise HTTPException(status_code=404, detail="Project not found")

    return result


@router.post("/api/project/{video_id}/extract/insights")
async def extract_insights_only(video_id: str):
    """Extract only insights from a project."""
    from ..services.extraction import run_extraction_and_save

    result = await run_extraction_and_save(
        video_id=video_id,
        extract_entities_flag=False,
        extract_insights_flag=True,
        extract_quotes_flag=False,
        extract_facts_flag=False,
        extract_sections_flag=False
    )

    if "error" in result and result["error"] == "Project not found":
        raise HTTPException(status_code=404, detail="Project not found")

    return result


@router.get("/api/project/{video_id}/entities")
async def get_entities(video_id: str):
    """Get extracted entities for a project."""
    project = get_project(video_id)
    if not project:
        raise HTTPException(status_code=404, detail="Project not found")

    return {
        "entities": project.entities.to_dict(),
        "total": project.entities.count()
    }


@router.get("/api/project/{video_id}/insights")
async def get_insights(video_id: str):
    """Get extracted insights for a project."""
    project = get_project(video_id)
    if not project:
        raise HTTPException(status_code=404, detail="Project not found")

    return {
        "insights": [i.to_dict() for i in project.insights],
        "total": len(project.insights)
    }


@router.get("/api/project/{video_id}/quotes")
async def get_quotes(video_id: str):
    """Get extracted quotes for a project."""
    project = get_project(video_id)
    if not project:
        raise HTTPException(status_code=404, detail="Project not found")

    return {
        "quotes": [q.to_dict() for q in project.quotes],
        "total": len(project.quotes)
    }


@router.get("/api/project/{video_id}/facts")
async def get_facts(video_id: str):
    """Get extracted facts for a project."""
    project = get_project(video_id)
    if not project:
        raise HTTPException(status_code=404, detail="Project not found")

    return {
        "facts": [f.to_dict() for f in project.facts],
        "total": len(project.facts)
    }


@router.get("/api/project/{video_id}/sections")
async def get_sections(video_id: str):
    """Get semantic sections for a project."""
    project = get_project(video_id)
    if not project:
        raise HTTPException(status_code=404, detail="Project not found")

    return {
        "sections": [s.to_dict() for s in project.sections],
        "total": len(project.sections)
    }



"""
Extraction Pipeline for Whisper Scribe 3.0.

Uses AI to extract entities, insights, quotes, facts, and semantic sections
from video transcriptions.
"""

import json
import uuid
import asyncio
from typing import List, Optional, Dict, Any
from datetime import datetime

from .eden_ai import chat_with_eden, calculate_chat_cost, estimate_tokens
from .project import (
    ProjectData, Entity, EntityType, Mention,
    ExtractedEntities, ExtractedInsight, InsightCategory,
    ExtractedQuote, ExtractedFact, FactCategory,
    SemanticSection, get_project, save_project
)
from .knowledge_graph import update_graph_for_project
from ..config import get_api_key, get_system_prompt


# =============================================================================
# Extraction Functions
# =============================================================================

async def extract_entities(
    transcription: str,
    api_key: str,
    provider: str = "openai"
) -> tuple[ExtractedEntities, dict]:
    """
    Extract named entities from transcription.

    Returns:
        Tuple of (ExtractedEntities, cost_info)
    """
    # Truncate if too long (keep first 12000 chars for entity extraction)
    text = transcription[:12000] if len(transcription) > 12000 else transcription

    # Get prompt from config (user-configurable)
    system_prompt = get_system_prompt("extraction_entities")
    prompt = f"""{system_prompt}

TRANSCRIPTION:
{text}"""

    messages = [{"role": "user", "content": prompt}]

    try:
        result = await chat_with_eden(
            messages=messages,
            api_key=api_key,
            providers=[provider],
            temperature=0.3,  # Lower temp for more consistent extraction
            max_tokens=3000
        )

        # Parse response
        generated = ""
        if provider in result and isinstance(result[provider], dict):
            generated = result[provider].get("generated_text", "")

        # Extract JSON from response
        entities_data = _parse_json_response(generated)

        # Convert to ExtractedEntities
        entities = ExtractedEntities(
            people=[_parse_entity(e, EntityType.PERSON) for e in entities_data.get("people", [])],
            organizations=[_parse_entity(e, EntityType.ORGANIZATION) for e in entities_data.get("organizations", [])],
            technologies=[_parse_entity(e, EntityType.TECHNOLOGY) for e in entities_data.get("technologies", [])],
            products=[_parse_entity(e, EntityType.PRODUCT) for e in entities_data.get("products", [])],
            places=[_parse_entity(e, EntityType.PLACE) for e in entities_data.get("places", [])],
            concepts=[_parse_entity(e, EntityType.CONCEPT) for e in entities_data.get("concepts", [])]
        )

        cost_info = result.get("cost_info", {})
        cost_info["operation"] = "entity_extraction"

        return entities, cost_info

    except Exception as e:
        # Return empty on error
        return ExtractedEntities(), {"error": str(e)}


async def extract_insights(
    transcription: str,
    api_key: str,
    provider: str = "openai"
) -> tuple[List[ExtractedInsight], dict]:
    """
    Extract key insights from transcription.

    Returns:
        Tuple of (list of insights, cost_info)
    """
    text = transcription[:12000] if len(transcription) > 12000 else transcription

    # Get prompt from config (user-configurable)
    system_prompt = get_system_prompt("extraction_insights")
    prompt = f"""{system_prompt}

TRANSCRIPTION:
{text}"""

    messages = [{"role": "user", "content": prompt}]

    try:
        result = await chat_with_eden(
            messages=messages,
            api_key=api_key,
            providers=[provider],
            temperature=0.5,
            max_tokens=2000
        )

        generated = ""
        if provider in result and isinstance(result[provider], dict):
            generated = result[provider].get("generated_text", "")

        insights_data = _parse_json_response(generated)

        insights = []
        for item in insights_data.get("insights", []):
            category = _parse_insight_category(item.get("category", "takeaway"))
            insights.append(ExtractedInsight(
                id=str(uuid.uuid4()),
                content=item.get("content", ""),
                category=category,
                confidence=0.9
            ))

        cost_info = result.get("cost_info", {})
        cost_info["operation"] = "insight_extraction"

        return insights, cost_info

    except Exception as e:
        return [], {"error": str(e)}


async def extract_quotes(
    transcription: str,
    api_key: str,
    provider: str = "openai"
) -> tuple[List[ExtractedQuote], dict]:
    """
    Extract notable quotes from transcription.

    Returns:
        Tuple of (list of quotes, cost_info)
    """
    text = transcription[:12000] if len(transcription) > 12000 else transcription

    # Get prompt from config (user-configurable)
    system_prompt = get_system_prompt("extraction_quotes")
    prompt = f"""{system_prompt}

TRANSCRIPTION:
{text}"""

    messages = [{"role": "user", "content": prompt}]

    try:
        result = await chat_with_eden(
            messages=messages,
            api_key=api_key,
            providers=[provider],
            temperature=0.4,
            max_tokens=2000
        )

        generated = ""
        if provider in result and isinstance(result[provider], dict):
            generated = result[provider].get("generated_text", "")

        quotes_data = _parse_json_response(generated)

        quotes = []
        for item in quotes_data.get("quotes", []):
            quotes.append(ExtractedQuote(
                id=str(uuid.uuid4()),
                text=item.get("text", ""),
                speaker=item.get("speaker"),
                context=item.get("context", "")
            ))

        cost_info = result.get("cost_info", {})
        cost_info["operation"] = "quote_extraction"

        return quotes, cost_info

    except Exception as e:
        return [], {"error": str(e)}


async def extract_facts(
    transcription: str,
    api_key: str,
    provider: str = "openai"
) -> tuple[List[ExtractedFact], dict]:
    """
    Extract facts and statistics from transcription.

    Returns:
        Tuple of (list of facts, cost_info)
    """
    text = transcription[:12000] if len(transcription) > 12000 else transcription

    # Get prompt from config (user-configurable)
    system_prompt = get_system_prompt("extraction_facts")
    prompt = f"""{system_prompt}

TRANSCRIPTION:
{text}"""

    messages = [{"role": "user", "content": prompt}]

    try:
        result = await chat_with_eden(
            messages=messages,
            api_key=api_key,
            providers=[provider],
            temperature=0.3,
            max_tokens=2000
        )

        generated = ""
        if provider in result and isinstance(result[provider], dict):
            generated = result[provider].get("generated_text", "")

        facts_data = _parse_json_response(generated)

        facts = []
        for item in facts_data.get("facts", []):
            category = _parse_fact_category(item.get("category", "claim"))
            facts.append(ExtractedFact(
                id=str(uuid.uuid4()),
                statement=item.get("statement", ""),
                category=category,
                source=item.get("source"),
                confidence=0.85
            ))

        cost_info = result.get("cost_info", {})
        cost_info["operation"] = "fact_extraction"

        return facts, cost_info

    except Exception as e:
        return [], {"error": str(e)}


async def extract_sections(
    transcription: str,
    api_key: str,
    provider: str = "openai",
    duration_seconds: int = 0
) -> tuple[List[SemanticSection], dict]:
    """
    Extract semantic sections from transcription.

    Returns:
        Tuple of (list of sections, cost_info)
    """
    text = transcription[:15000] if len(transcription) > 15000 else transcription

    # Get prompt from config (user-configurable)
    system_prompt = get_system_prompt("extraction_sections")
    prompt = f"""{system_prompt}

TRANSCRIPTION:
{text}"""

    messages = [{"role": "user", "content": prompt}]

    try:
        result = await chat_with_eden(
            messages=messages,
            api_key=api_key,
            providers=[provider],
            temperature=0.4,
            max_tokens=2000
        )

        generated = ""
        if provider in result and isinstance(result[provider], dict):
            generated = result[provider].get("generated_text", "")

        sections_data = _parse_json_response(generated)

        sections = []
        raw_sections = sections_data.get("sections", [])
        num_sections = len(raw_sections)

        for i, item in enumerate(raw_sections):
            # Estimate timestamps based on position
            position = item.get("approximate_position", "middle")
            start_time, end_time = _estimate_timestamps(
                position, i, num_sections, duration_seconds
            )

            sections.append(SemanticSection(
                id=str(uuid.uuid4()),
                title=item.get("title", f"Section {i+1}"),
                content="",  # Content can be filled from transcription later
                start_time=start_time,
                end_time=end_time,
                topics=item.get("topics", []),
                summary=item.get("summary", "")
            ))

        cost_info = result.get("cost_info", {})
        cost_info["operation"] = "section_extraction"

        return sections, cost_info

    except Exception as e:
        return [], {"error": str(e)}


# =============================================================================
# Extraction Orchestrator
# =============================================================================

async def run_full_extraction(
    video_id: str,
    transcription: str,
    api_key: str,
    provider: str = "openai",
    duration_seconds: int = 0,
    extract_entities_flag: bool = True,
    extract_insights_flag: bool = True,
    extract_quotes_flag: bool = True,
    extract_facts_flag: bool = True,
    extract_sections_flag: bool = True
) -> Dict[str, Any]:
    """
    Run the full extraction pipeline on a transcription.

    Args:
        video_id: The video/project ID
        transcription: The full transcription text
        api_key: Eden.ai API key
        provider: AI provider to use
        duration_seconds: Video duration for timestamp estimation
        extract_*_flag: Flags to control which extractions to run

    Returns:
        Dict with extraction results and costs
    """
    results = {
        "video_id": video_id,
        "status": "in_progress",
        "entities": None,
        "insights": [],
        "quotes": [],
        "facts": [],
        "sections": [],
        "costs": [],
        "total_cost": 0.0,
        "errors": []
    }

    # Run extractions in parallel where possible
    tasks = []

    if extract_entities_flag:
        tasks.append(("entities", extract_entities(transcription, api_key, provider)))

    if extract_insights_flag:
        tasks.append(("insights", extract_insights(transcription, api_key, provider)))

    if extract_quotes_flag:
        tasks.append(("quotes", extract_quotes(transcription, api_key, provider)))

    if extract_facts_flag:
        tasks.append(("facts", extract_facts(transcription, api_key, provider)))

    if extract_sections_flag:
        tasks.append(("sections", extract_sections(transcription, api_key, provider, duration_seconds)))

    # Execute all tasks
    if tasks:
        task_results = await asyncio.gather(
            *[task[1] for task in tasks],
            return_exceptions=True
        )

        # Process results
        total_cost = 0.0
        for i, (name, _) in enumerate(tasks):
            task_result = task_results[i]

            if isinstance(task_result, Exception):
                results["errors"].append({
                    "extraction": name,
                    "error": str(task_result)
                })
            else:
                data, cost_info = task_result
                results[name] = data if name == "entities" else data

                if cost_info and "cost_usd" in cost_info:
                    results["costs"].append(cost_info)
                    total_cost += cost_info.get("cost_usd", 0)
                elif cost_info and "error" in cost_info:
                    results["errors"].append({
                        "extraction": name,
                        "error": cost_info["error"]
                    })

        results["total_cost"] = round(total_cost, 6)

    results["status"] = "complete" if not results["errors"] else "partial"

    return results


async def run_extraction_and_save(
    video_id: str,
    api_key: str = None,
    provider: str = "openai",
    **kwargs
) -> Dict[str, Any]:
    """
    Run extraction pipeline and save results to project.

    Args:
        video_id: The project/video ID
        api_key: Eden.ai API key (uses config if not provided)
        provider: AI provider to use
        **kwargs: Additional flags for run_full_extraction

    Returns:
        Extraction results dict
    """
    # Get project
    project = get_project(video_id)
    if not project:
        return {"error": "Project not found", "video_id": video_id}

    # Get API key
    if not api_key:
        api_key = get_api_key("eden")
        if not api_key:
            return {"error": "No Eden.ai API key configured", "video_id": video_id}

    # Update status
    project.extraction_status = "in_progress"
    save_project(project)

    try:
        # Run extraction
        results = await run_full_extraction(
            video_id=video_id,
            transcription=project.transcription,
            api_key=api_key,
            provider=provider,
            duration_seconds=project.duration,
            **kwargs
        )

        # Update project with results
        if results.get("entities"):
            project.entities = results["entities"]

        if results.get("insights"):
            project.insights = results["insights"]

        if results.get("quotes"):
            project.quotes = results["quotes"]

        if results.get("facts"):
            project.facts = results["facts"]

        if results.get("sections"):
            project.sections = results["sections"]

        # Update costs
        for cost in results.get("costs", []):
            project.costs["entries"].append(cost)
            project.costs["total_cost"] += cost.get("cost_usd", 0)
        project.costs["total_cost"] = round(project.costs["total_cost"], 6)

        # Extract topics from sections and entities
        topics = set(project.shared_topics)
        for section in project.sections:
            topics.update(section.topics)
        for entity in project.entities.technologies:
            topics.add(entity.name)
        for entity in project.entities.concepts:
            topics.add(entity.name)
        project.shared_topics = list(topics)

        # Update status
        project.extraction_status = "complete" if results["status"] == "complete" else "partial"
        project.processed_at = datetime.now().isoformat()

        # Save project
        save_project(project)

        # Update knowledge graph
        update_graph_for_project(project)

        return results

    except Exception as e:
        project.extraction_status = "error"
        save_project(project)
        return {"error": str(e), "video_id": video_id}


# =============================================================================
# Helper Functions
# =============================================================================

def _parse_json_response(text: str) -> dict:
    """Parse JSON from AI response, handling common issues."""
    if not text:
        return {}

    # Try to find JSON in the response
    text = text.strip()

    # Remove markdown code blocks if present
    if text.startswith("```json"):
        text = text[7:]
    elif text.startswith("```"):
        text = text[3:]
    if text.endswith("```"):
        text = text[:-3]

    text = text.strip()

    try:
        return json.loads(text)
    except json.JSONDecodeError:
        # Try to find JSON object in text
        start = text.find("{")
        end = text.rfind("}") + 1
        if start >= 0 and end > start:
            try:
                return json.loads(text[start:end])
            except json.JSONDecodeError:
                pass
        return {}


def _parse_entity(data: dict, entity_type: EntityType) -> Entity:
    """Parse entity data into Entity object."""
    mentions = []
    for mention_text in data.get("mentions", []):
        if isinstance(mention_text, str):
            mentions.append(Mention(timestamp=0.0, context=mention_text))
        elif isinstance(mention_text, dict):
            mentions.append(Mention(
                timestamp=mention_text.get("timestamp", 0.0),
                context=mention_text.get("context", "")
            ))

    return Entity(
        id=str(uuid.uuid4()),
        name=data.get("name", "Unknown"),
        type=entity_type,
        description=data.get("description", ""),
        mentions=mentions,
        wikipedia_url=data.get("wikipedia_url")
    )


def _parse_insight_category(category_str: str) -> InsightCategory:
    """Parse insight category string to enum."""
    category_map = {
        "takeaway": InsightCategory.TAKEAWAY,
        "aha": InsightCategory.AHA,
        "wisdom": InsightCategory.WISDOM,
        "action": InsightCategory.ACTION,
        "warning": InsightCategory.WARNING,
        "definition": InsightCategory.DEFINITION
    }
    return category_map.get(category_str.lower(), InsightCategory.TAKEAWAY)


def _parse_fact_category(category_str: str) -> FactCategory:
    """Parse fact category string to enum."""
    category_map = {
        "statistic": FactCategory.STATISTIC,
        "definition": FactCategory.DEFINITION,
        "claim": FactCategory.CLAIM,
        "date": FactCategory.DATE,
        "comparison": FactCategory.COMPARISON
    }
    return category_map.get(category_str.lower(), FactCategory.CLAIM)


def _estimate_timestamps(
    position: str,
    index: int,
    total_sections: int,
    duration_seconds: int
) -> tuple[float, float]:
    """Estimate start and end timestamps for a section."""
    if duration_seconds <= 0:
        return 0.0, 0.0

    # Use position hint and index to estimate
    position_map = {
        "beginning": 0.0,
        "early": 0.2,
        "middle": 0.5,
        "late": 0.7,
        "end": 0.9
    }

    # Calculate based on index if we have sections
    if total_sections > 1:
        section_duration = duration_seconds / total_sections
        start_time = index * section_duration
        end_time = (index + 1) * section_duration
    else:
        # Use position hint
        pos_fraction = position_map.get(position.lower(), 0.5)
        section_duration = duration_seconds * 0.2  # Assume ~20% of video
        start_time = pos_fraction * duration_seconds
        end_time = min(start_time + section_duration, duration_seconds)

    return round(start_time, 1), round(end_time, 1)


# =============================================================================
# Sync Wrappers
# =============================================================================

def run_extraction_sync(video_id: str, **kwargs) -> Dict[str, Any]:
    """Synchronous wrapper for run_extraction_and_save."""
    import asyncio

    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # We're in an async context, create a new task
            import concurrent.futures
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(
                    asyncio.run,
                    run_extraction_and_save(video_id, **kwargs)
                )
                return future.result()
        else:
            return loop.run_until_complete(run_extraction_and_save(video_id, **kwargs))
    except RuntimeError:
        # No event loop, create one
        return asyncio.run(run_extraction_and_save(video_id, **kwargs))



"""Multi-provider AI service router - all requests go through Eden.ai."""

from typing import Optional, Dict
import asyncio
from ..config import load_config, get_eden_api_key, get_provider_settings, get_system_prompt


def _run_async(coro):
    """Helper to run async code synchronously, handling nested loops."""
    try:
        # Check if there's already a running loop
        loop = asyncio.get_running_loop()
        # If we're in an async context, use nest_asyncio or run in thread
        import concurrent.futures
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(asyncio.run, coro)
            return future.result(timeout=120)
    except RuntimeError:
        # No running loop, safe to create one
        return asyncio.run(coro)


def get_eden_key_or_error():
    """Get Eden API key or raise descriptive error."""
    api_key = get_eden_api_key()
    if not api_key:
        raise ValueError("Eden.ai API key not configured. Set it in Settings → Eden.ai tab.")
    return api_key


def analyze_with_provider(
    text: str,
    analysis_type: str = "comprehensive",
    provider: str = None
) -> dict:
    """
    Analyze text using Eden.ai with the configured provider.

    Args:
        text: Text to analyze
        analysis_type: Type of analysis
        provider: Override provider (uses settings if None)

    Returns:
        Analysis results including topics
    """
    api_key = get_eden_key_or_error()

    if provider is None:
        settings = get_provider_settings()
        provider = settings.get("analysis_provider", "openai")

    from .eden_ai import chat_with_eden

    # Map provider names to Eden provider strings
    eden_provider = _map_to_eden_provider(provider)

    # Get the system prompt from config (user-configurable)
    system_prompt = get_system_prompt("analysis_full")

    # First, get the main analysis
    messages = [{
        "role": "user",
        "content": f"""{system_prompt}

--- TRANSCRIPTION START ---
{text[:8000]}
--- TRANSCRIPTION END ---"""
    }]

    result = _run_async(chat_with_eden(messages, api_key, providers=[eden_provider]))

    # Extract response from Eden format
    analysis_text = _extract_eden_response(result)

    # Now extract topics separately for better tagging
    topics = _extract_topics(text, api_key, eden_provider)

    return {
        "analysis": analysis_text,
        "topics": topics,
        "provider": provider,
        "via": "eden",
        "cost_info": result.get("cost_info")
    }


def _extract_topics(text: str, api_key: str, provider: str) -> list:
    """
    Extract topics/tags from text.
    Returns a list of 3-7 topic strings.
    """
    from .eden_ai import chat_with_eden

    # Get the system prompt from config (user-configurable)
    system_prompt = get_system_prompt("analysis_topics")

    messages = [{
        "role": "user",
        "content": f"""{system_prompt}

Transcription:
{text[:4000]}"""
    }]

    try:
        result = _run_async(chat_with_eden(messages, api_key, providers=[provider], max_tokens=200))
        response_text = _extract_eden_response(result)

        # Parse topics from response
        topics = []
        for line in response_text.strip().split('\n'):
            line = line.strip()
            # Remove any numbering, bullets, or dashes
            line = line.lstrip('0123456789.-•*) ').strip()
            if line and len(line) < 50:  # Reasonable topic length
                topics.append(line)

        return topics[:7]  # Max 7 topics
    except Exception as e:
        print(f"Topics extraction failed: {e}")
        return []


def critique_with_provider(
    text: str,
    original_analysis: str,
    provider: str = None
) -> str:
    """
    Critique an analysis using Eden.ai with the configured provider.

    Args:
        text: Original transcription
        original_analysis: The analysis to critique
        provider: Override provider (uses settings if None)

    Returns:
        Critique text
    """
    api_key = get_eden_key_or_error()

    if provider is None:
        settings = get_provider_settings()
        provider = settings.get("critique_provider", "anthropic")

    from .eden_ai import chat_with_eden

    eden_provider = _map_to_eden_provider(provider)

    # Get the system prompt from config (user-configurable)
    system_prompt = get_system_prompt("critique_main")

    messages = [{
        "role": "user",
        "content": f"""{system_prompt}

Original Transcription (excerpt):
--- TRANSCRIPTION START ---
{text[:3000]}
--- TRANSCRIPTION END ---

Analysis to critique:
{original_analysis}"""
    }]

    result = _run_async(chat_with_eden(messages, api_key, providers=[eden_provider]))
    return _extract_eden_response(result)


def find_resources_with_provider(
    topic: str,
    context: str = None,
    provider: str = None
) -> dict:
    """
    Find external resources using Eden.ai with the configured provider.

    Args:
        topic: Topic to search for
        context: Optional context from transcription
        provider: Override provider (uses settings if None)

    Returns:
        Resources and references with citations
    """
    api_key = get_eden_key_or_error()

    if provider is None:
        settings = get_provider_settings()
        provider = settings.get("resources_provider", "perplexityai")

    # Use specialized Perplexity search for citations
    if provider == "perplexity" or provider == "perplexityai":
        from .eden_ai import search_with_perplexity

        # Get the system prompt from config (user-configurable)
        system_prompt = get_system_prompt("resources_search")

        query = f"""{system_prompt}

Topic: {topic}"""

        result = _run_async(search_with_perplexity(query, api_key, context))

        return {
            "resources": result.get("answer", ""),
            "citations": result.get("citations", []),
            "provider": "perplexityai",
            "via": "eden",
            "topic": topic
        }

    # For other providers, use chat
    from .eden_ai import chat_with_eden

    eden_provider = _map_to_eden_provider(provider)

    # Get the system prompt from config (user-configurable)
    system_prompt = get_system_prompt("resources_search")

    query = f"{system_prompt}\n\nTopic: {topic}"
    if context:
        query += f"\n\nContext from the content:\n{context[:1500]}"

    messages = [{"role": "user", "content": query}]

    result = _run_async(chat_with_eden(messages, api_key, providers=[eden_provider]))
    response_text = _extract_eden_response(result)

    return {
        "resources": response_text,
        "provider": provider,
        "via": "eden",
        "topic": topic
    }


def answer_question_with_provider(
    text: str,
    question: str,
    provider: str = None
) -> dict:
    """
    Answer a question using Eden.ai with the configured provider.

    Args:
        text: Transcription text
        question: User's question
        provider: Override provider (uses settings if None)

    Returns:
        Answer and metadata
    """
    api_key = get_eden_key_or_error()

    if provider is None:
        settings = get_provider_settings()
        provider = settings.get("analysis_provider", "openai")

    from .eden_ai import chat_with_eden

    eden_provider = _map_to_eden_provider(provider)

    # Get the system prompt from config (user-configurable)
    system_prompt = get_system_prompt("qa_answer")

    messages = [{
        "role": "user",
        "content": f"""{system_prompt}

--- TRANSCRIPTION START ---
{text[:8000]}
--- TRANSCRIPTION END ---

Question: {question}"""
    }]

    result = _run_async(chat_with_eden(messages, api_key, providers=[eden_provider]))
    answer = _extract_eden_response(result)

    return {"answer": answer, "provider": provider, "via": "eden"}


def comprehensive_qa(
    text: str,
    question: str,
    find_references: bool = True
) -> dict:
    """
    Comprehensive Q&A with optional critique and references.
    All requests go through Eden.ai.

    Args:
        text: Transcription text
        question: User's question
        find_references: Whether to find external references

    Returns:
        Complete Q&A response with all provider results
    """
    settings = get_provider_settings()
    result = {}

    # 1. Get initial answer from analysis provider
    try:
        answer_result = answer_question_with_provider(text, question)
        result["initial_answer"] = answer_result
    except Exception as e:
        result["initial_answer_error"] = str(e)

    # 2. Get critique from critique provider (if different from analysis)
    if settings.get("critique_provider") != settings.get("analysis_provider"):
        try:
            initial_answer = result.get("initial_answer", {}).get("answer", "")
            if initial_answer:
                critique = critique_with_provider(text, initial_answer)
                result["critique"] = critique
        except Exception as e:
            result["critique_error"] = str(e)

    # 3. Find external references if requested
    if find_references:
        try:
            resources = find_resources_with_provider(question, text[:2000])
            result["references"] = resources
        except Exception as e:
            result["references_error"] = str(e)

    return result


def _map_to_eden_provider(provider: str) -> str:
    """Map our provider names to Eden.ai provider strings."""
    mapping = {
        # Analysis/Chat providers
        "openai": "openai",
        "anthropic": "anthropic",
        "google": "google",
        "cohere": "cohere",
        "meta": "meta",
        "perplexity": "perplexityai",  # Perplexity for web search & citations
        "perplexityai": "perplexityai",
        # Transcription providers
        "assemblyai": "assemblyai",
        "deepgram": "deepgram",
        "amazon": "amazon",
        "microsoft": "microsoft",
    }
    return mapping.get(provider, "openai")


def _extract_eden_response(result: dict) -> str:
    """Extract the text response from Eden.ai result format."""
    if not result:
        return "No response received"

    # Eden returns results keyed by provider name
    for provider_name, provider_result in result.items():
        if isinstance(provider_result, dict):
            # Try different response formats
            if 'generated_text' in provider_result:
                return provider_result['generated_text']
            if 'message' in provider_result:
                return provider_result['message']
            if 'text' in provider_result:
                return provider_result['text']

    # Fallback: return string representation
    return str(result)



"""
Knowledge Graph service for Whisper Scribe 3.0.

Builds and maintains a knowledge graph connecting videos, topics, entities,
and concepts for visualization and relationship discovery.
"""

import json
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Dict, Set, Tuple
from dataclasses import dataclass, field
from collections import defaultdict
from ..config import get_config_dir
from .project import load_projects, ProjectData


@dataclass
class GraphNode:
    """A node in the knowledge graph."""
    id: str
    type: str  # "video", "topic", "person", "organization", "technology", "concept"
    label: str
    metadata: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'type': self.type,
            'label': self.label,
            'metadata': self.metadata
        }


@dataclass
class GraphEdge:
    """An edge connecting two nodes."""
    source: str  # Node ID
    target: str  # Node ID
    type: str    # "mentions", "related_to", "part_of", "shares_topic"
    weight: float = 1.0
    metadata: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'source': self.source,
            'target': self.target,
            'type': self.type,
            'weight': self.weight,
            'metadata': self.metadata
        }


@dataclass
class KnowledgeGraph:
    """The complete knowledge graph."""
    nodes: List[GraphNode] = field(default_factory=list)
    edges: List[GraphEdge] = field(default_factory=list)
    last_updated: str = ""

    def __post_init__(self):
        if not self.last_updated:
            self.last_updated = datetime.now().isoformat()

    def to_dict(self) -> dict:
        return {
            'nodes': [n.to_dict() for n in self.nodes],
            'edges': [e.to_dict() for e in self.edges],
            'last_updated': self.last_updated
        }

    @classmethod
    def from_dict(cls, data: dict) -> 'KnowledgeGraph':
        return cls(
            nodes=[GraphNode(**n) for n in data.get('nodes', [])],
            edges=[GraphEdge(**e) for e in data.get('edges', [])],
            last_updated=data.get('last_updated', '')
        )

    def add_node(self, node: GraphNode) -> None:
        """Add a node if it doesn't already exist."""
        if not any(n.id == node.id for n in self.nodes):
            self.nodes.append(node)

    def add_edge(self, edge: GraphEdge) -> None:
        """Add an edge, updating weight if it already exists."""
        for e in self.edges:
            if e.source == edge.source and e.target == edge.target and e.type == edge.type:
                e.weight = max(e.weight, edge.weight)
                return
        self.edges.append(edge)

    def get_node(self, node_id: str) -> Optional[GraphNode]:
        """Get a node by ID."""
        for n in self.nodes:
            if n.id == node_id:
                return n
        return None

    def get_connected_nodes(self, node_id: str) -> List[Tuple[GraphNode, GraphEdge]]:
        """Get all nodes connected to a given node."""
        connected = []
        for edge in self.edges:
            if edge.source == node_id:
                node = self.get_node(edge.target)
                if node:
                    connected.append((node, edge))
            elif edge.target == node_id:
                node = self.get_node(edge.source)
                if node:
                    connected.append((node, edge))
        return connected

    def get_nodes_by_type(self, node_type: str) -> List[GraphNode]:
        """Get all nodes of a specific type."""
        return [n for n in self.nodes if n.type == node_type]

    def get_video_connections(self, video_id: str) -> List[dict]:
        """Get all videos connected to a given video with connection details."""
        video_node_id = f"video:{video_id}"
        connections = []
        seen_videos = set()

        for node, edge in self.get_connected_nodes(video_node_id):
            if node.type == "video" and node.id not in seen_videos:
                seen_videos.add(node.id)
                connections.append({
                    'video_id': node.id.replace('video:', ''),
                    'title': node.label,
                    'connection_type': edge.type,
                    'weight': edge.weight
                })
            elif node.type in ("topic", "person", "organization", "technology"):
                # Find other videos connected to this entity
                for sub_node, sub_edge in self.get_connected_nodes(node.id):
                    if sub_node.type == "video" and sub_node.id != video_node_id and sub_node.id not in seen_videos:
                        seen_videos.add(sub_node.id)
                        connections.append({
                            'video_id': sub_node.id.replace('video:', ''),
                            'title': sub_node.label,
                            'via': node.label,
                            'via_type': node.type,
                            'connection_type': 'shared_entity',
                            'weight': min(edge.weight, sub_edge.weight)
                        })

        # Sort by weight descending
        connections.sort(key=lambda x: x.get('weight', 0), reverse=True)
        return connections


# =============================================================================
# Storage Functions
# =============================================================================

def get_graph_file() -> Path:
    """Get the knowledge graph file path."""
    return get_config_dir() / "knowledge_graph.json"


def load_graph() -> KnowledgeGraph:
    """Load the knowledge graph from file."""
    graph_file = get_graph_file()
    if not graph_file.exists():
        return KnowledgeGraph()

    try:
        with open(graph_file, 'r', encoding='utf-8') as f:
            data = json.load(f)
            return KnowledgeGraph.from_dict(data)
    except (json.JSONDecodeError, Exception):
        return KnowledgeGraph()


def save_graph(graph: KnowledgeGraph) -> None:
    """Save the knowledge graph to file."""
    graph.last_updated = datetime.now().isoformat()
    graph_file = get_graph_file()
    with open(graph_file, 'w', encoding='utf-8') as f:
        json.dump(graph.to_dict(), f, indent=2, ensure_ascii=False)


# =============================================================================
# Graph Building Functions
# =============================================================================

def build_graph_from_projects() -> KnowledgeGraph:
    """
    Build the complete knowledge graph from all projects.
    This rebuilds the graph from scratch.
    """
    graph = KnowledgeGraph()
    projects = load_projects()

    topic_videos: Dict[str, Set[str]] = defaultdict(set)  # topic -> set of video_ids
    entity_videos: Dict[str, Set[str]] = defaultdict(set)  # entity_key -> set of video_ids

    for p_data in projects:
        project = ProjectData.from_dict(p_data)

        # Add video node
        video_node_id = f"video:{project.video_id}"
        graph.add_node(GraphNode(
            id=video_node_id,
            type="video",
            label=project.title,
            metadata={
                'channel': project.channel,
                'duration': project.duration,
                'thumbnail': project.thumbnail,
                'url': project.url
            }
        ))

        # Process topics
        all_topics = set(project.shared_topics) | set(project.tags)

        # Also extract topics from analysis if available
        if project.analyses and project.primary_analysis_provider:
            primary_analysis = project.analyses.get(project.primary_analysis_provider, {})
            content = primary_analysis.get('content', {})
            if isinstance(content, dict):
                analysis_topics = content.get('topics', [])
                if isinstance(analysis_topics, list):
                    all_topics.update(analysis_topics)

        for topic in all_topics:
            if topic:
                topic_node_id = f"topic:{topic.lower()}"
                graph.add_node(GraphNode(
                    id=topic_node_id,
                    type="topic",
                    label=topic
                ))
                graph.add_edge(GraphEdge(
                    source=video_node_id,
                    target=topic_node_id,
                    type="has_topic",
                    weight=1.0
                ))
                topic_videos[topic.lower()].add(project.video_id)

        # Process entities
        for entity_list, entity_type in [
            (project.entities.people, "person"),
            (project.entities.organizations, "organization"),
            (project.entities.technologies, "technology"),
            (project.entities.products, "product"),
            (project.entities.places, "place"),
            (project.entities.concepts, "concept")
        ]:
            for entity in entity_list:
                entity_key = f"{entity_type}:{entity.name.lower()}"
                graph.add_node(GraphNode(
                    id=entity_key,
                    type=entity_type,
                    label=entity.name,
                    metadata={
                        'description': entity.description,
                        'wikipedia_url': entity.wikipedia_url
                    }
                ))
                graph.add_edge(GraphEdge(
                    source=video_node_id,
                    target=entity_key,
                    type="mentions",
                    weight=len(entity.mentions) if entity.mentions else 1.0
                ))
                entity_videos[entity_key].add(project.video_id)

    # Add edges between videos that share topics
    for topic, videos in topic_videos.items():
        video_list = list(videos)
        for i, v1 in enumerate(video_list):
            for v2 in video_list[i + 1:]:
                graph.add_edge(GraphEdge(
                    source=f"video:{v1}",
                    target=f"video:{v2}",
                    type="shares_topic",
                    weight=1.0,
                    metadata={'topic': topic}
                ))

    # Add edges between videos that share entities
    for entity_key, videos in entity_videos.items():
        video_list = list(videos)
        for i, v1 in enumerate(video_list):
            for v2 in video_list[i + 1:]:
                graph.add_edge(GraphEdge(
                    source=f"video:{v1}",
                    target=f"video:{v2}",
                    type="shares_entity",
                    weight=1.5,  # Higher weight for shared entities
                    metadata={'entity': entity_key}
                ))

    save_graph(graph)
    return graph


def update_graph_for_project(project: ProjectData) -> KnowledgeGraph:
    """
    Update the knowledge graph for a single project.
    More efficient than rebuilding the entire graph.
    """
    graph = load_graph()

    video_node_id = f"video:{project.video_id}"

    # Remove existing edges for this video
    graph.edges = [e for e in graph.edges if e.source != video_node_id and e.target != video_node_id]

    # Add/update video node
    existing_node = graph.get_node(video_node_id)
    if existing_node:
        existing_node.label = project.title
        existing_node.metadata = {
            'channel': project.channel,
            'duration': project.duration,
            'thumbnail': project.thumbnail,
            'url': project.url
        }
    else:
        graph.add_node(GraphNode(
            id=video_node_id,
            type="video",
            label=project.title,
            metadata={
                'channel': project.channel,
                'duration': project.duration,
                'thumbnail': project.thumbnail,
                'url': project.url
            }
        ))

    # Add topic connections
    all_topics = set(project.shared_topics) | set(project.tags)
    for topic in all_topics:
        if topic:
            topic_node_id = f"topic:{topic.lower()}"
            graph.add_node(GraphNode(
                id=topic_node_id,
                type="topic",
                label=topic
            ))
            graph.add_edge(GraphEdge(
                source=video_node_id,
                target=topic_node_id,
                type="has_topic",
                weight=1.0
            ))

    # Add entity connections
    for entity_list, entity_type in [
        (project.entities.people, "person"),
        (project.entities.organizations, "organization"),
        (project.entities.technologies, "technology"),
        (project.entities.products, "product"),
        (project.entities.places, "place"),
        (project.entities.concepts, "concept")
    ]:
        for entity in entity_list:
            entity_key = f"{entity_type}:{entity.name.lower()}"
            graph.add_node(GraphNode(
                id=entity_key,
                type=entity_type,
                label=entity.name,
                metadata={
                    'description': entity.description,
                    'wikipedia_url': entity.wikipedia_url
                }
            ))
            graph.add_edge(GraphEdge(
                source=video_node_id,
                target=entity_key,
                type="mentions",
                weight=len(entity.mentions) if entity.mentions else 1.0
            ))

    # Find connections to other videos through shared topics/entities
    for node in graph.nodes:
        if node.type == "video" and node.id != video_node_id:
            # Check for shared topics
            other_video_topics = set()
            for edge in graph.edges:
                if edge.source == node.id and edge.type == "has_topic":
                    other_video_topics.add(edge.target)

            this_video_topics = {f"topic:{t.lower()}" for t in all_topics if t}
            shared = other_video_topics & this_video_topics
            if shared:
                graph.add_edge(GraphEdge(
                    source=video_node_id,
                    target=node.id,
                    type="shares_topic",
                    weight=len(shared)
                ))

    save_graph(graph)
    return graph


def remove_video_from_graph(video_id: str) -> KnowledgeGraph:
    """Remove a video and its connections from the graph."""
    graph = load_graph()

    video_node_id = f"video:{video_id}"

    # Remove the video node
    graph.nodes = [n for n in graph.nodes if n.id != video_node_id]

    # Remove edges involving this video
    graph.edges = [e for e in graph.edges if e.source != video_node_id and e.target != video_node_id]

    # Clean up orphaned nodes (topics/entities with no video connections)
    connected_nodes = set()
    for edge in graph.edges:
        connected_nodes.add(edge.source)
        connected_nodes.add(edge.target)

    graph.nodes = [n for n in graph.nodes if n.type == "video" or n.id in connected_nodes]

    save_graph(graph)
    return graph


# =============================================================================
# Query Functions
# =============================================================================

def get_graph_stats() -> dict:
    """Get statistics about the knowledge graph."""
    graph = load_graph()

    node_counts = defaultdict(int)
    for node in graph.nodes:
        node_counts[node.type] += 1

    edge_counts = defaultdict(int)
    for edge in graph.edges:
        edge_counts[edge.type] += 1

    return {
        'total_nodes': len(graph.nodes),
        'total_edges': len(graph.edges),
        'node_counts': dict(node_counts),
        'edge_counts': dict(edge_counts),
        'last_updated': graph.last_updated
    }


def get_topic_clusters() -> List[dict]:
    """Get topics clustered with their related videos and entities."""
    graph = load_graph()

    topics = []
    for node in graph.nodes:
        if node.type == "topic":
            connected = graph.get_connected_nodes(node.id)
            videos = []
            entities = []

            for conn_node, edge in connected:
                if conn_node.type == "video":
                    videos.append({
                        'video_id': conn_node.id.replace('video:', ''),
                        'title': conn_node.label
                    })
                elif conn_node.type in ("person", "organization", "technology"):
                    entities.append({
                        'name': conn_node.label,
                        'type': conn_node.type
                    })

            topics.append({
                'topic': node.label,
                'video_count': len(videos),
                'videos': videos,
                'related_entities': entities
            })

    # Sort by video count descending
    topics.sort(key=lambda x: x['video_count'], reverse=True)
    return topics


def find_related_videos(video_id: str, limit: int = 10) -> List[dict]:
    """Find videos related to a given video through shared topics/entities."""
    graph = load_graph()
    return graph.get_video_connections(video_id)[:limit]


def get_entity_network(entity_type: str = None) -> dict:
    """Get the network of entities, optionally filtered by type."""
    graph = load_graph()

    nodes = []
    edges = []

    for node in graph.nodes:
        if entity_type is None or node.type == entity_type:
            if node.type != "video":
                nodes.append(node.to_dict())

    # Get edges between non-video nodes
    node_ids = {n['id'] for n in nodes}
    for edge in graph.edges:
        if edge.source in node_ids and edge.target in node_ids:
            edges.append(edge.to_dict())

    return {
        'nodes': nodes,
        'edges': edges
    }


def search_graph(query: str) -> List[dict]:
    """Search for nodes matching a query."""
    graph = load_graph()
    query_lower = query.lower()

    results = []
    for node in graph.nodes:
        score = 0
        if query_lower in node.label.lower():
            score = 1.0 if node.label.lower() == query_lower else 0.5
        elif query_lower in str(node.metadata).lower():
            score = 0.3

        if score > 0:
            connected = graph.get_connected_nodes(node.id)
            results.append({
                'node': node.to_dict(),
                'score': score,
                'connections': len(connected)
            })

    results.sort(key=lambda x: (x['score'], x['connections']), reverse=True)
    return results



"""History management service."""

import json
import os
from datetime import datetime
from pathlib import Path
from typing import List, Optional
from ..config import get_config_dir


def get_history_file() -> Path:
    """Get the history file path."""
    return get_config_dir() / "history.json"


def load_history() -> List[dict]:
    """Load history from file."""
    history_file = get_history_file()
    if not history_file.exists():
        return []

    try:
        with open(history_file, 'r', encoding='utf-8') as f:
            return json.load(f)
    except (json.JSONDecodeError, Exception):
        return []


def save_history(history: List[dict]) -> None:
    """Save history to file."""
    history_file = get_history_file()
    with open(history_file, 'w', encoding='utf-8') as f:
        json.dump(history, f, indent=2, ensure_ascii=False)


def add_to_history(
    id: str,
    title: str,
    url: str,
    file_path: str,
    duration: Optional[str] = None,
    thumbnail: Optional[str] = None,
    initial_cost: Optional[dict] = None,
    language: Optional[str] = None,
    text_direction: Optional[str] = None
) -> dict:
    """Add a new item to history."""
    history = load_history()

    item = {
        'id': id,
        'title': title,
        'url': url,
        'file_path': file_path,
        'created_at': datetime.now().isoformat(),
        'duration': duration,
        'thumbnail': thumbnail,
        'language': language,
        'text_direction': text_direction or 'ltr',
        'analysis': None,  # Will store analysis results
        'translations': {},  # Will store translations by language
        'qa_history': [],  # Q&A exchanges
        'resources': None,  # Saved resources/citations
        'costs': {
            'entries': [],
            'total_cost': 0.0
        }
    }

    # Add initial cost if provided (e.g., transcription cost)
    if initial_cost:
        item['costs']['entries'].append(initial_cost)
        item['costs']['total_cost'] = initial_cost.get('cost_usd', 0)

    # Add to beginning of list
    history.insert(0, item)

    # Keep only last 100 items
    history = history[:100]

    save_history(history)
    return item


def update_history_item(id: str, **kwargs) -> Optional[dict]:
    """Update a history item with new data."""
    history = load_history()

    for i, item in enumerate(history):
        if item.get('id') == id:
            item.update(kwargs)
            history[i] = item
            save_history(history)
            return item
    return None


def get_history(search: Optional[str] = None, limit: int = 50, offset: int = 0) -> tuple:
    """
    Get history items with optional search.
    Searches in title and transcription text content.

    Returns (items, total_count)
    """
    history = load_history()

    if search:
        search_lower = search.lower()
        filtered = []
        for item in history:
            # Search in title
            if search_lower in item.get('title', '').lower():
                filtered.append(item)
                continue
            # Search in transcription text
            file_path = item.get('file_path')
            if file_path:
                text = get_transcription_text(file_path)
                if text and search_lower in text.lower():
                    filtered.append(item)
        history = filtered

    total = len(history)
    items = history[offset:offset + limit]

    return items, total


def get_history_item(id: str) -> Optional[dict]:
    """Get a specific history item by ID."""
    history = load_history()
    for item in history:
        if item.get('id') == id:
            return item
    return None


def delete_history_item(id: str) -> bool:
    """Delete a history item by ID."""
    history = load_history()
    original_len = len(history)

    history = [item for item in history if item.get('id') != id]

    if len(history) < original_len:
        save_history(history)
        return True
    return False


def clear_history() -> None:
    """Clear all history."""
    save_history([])


def get_transcription_text(file_path: str) -> Optional[str]:
    """Read transcription text from file."""
    try:
        if os.path.exists(file_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
                # Skip header if present
                if content.startswith('Title:'):
                    lines = content.split('\n')
                    # Find the separator line and skip past it
                    for i, line in enumerate(lines):
                        if line.startswith('==='):
                            return '\n'.join(lines[i+2:]).strip()
                return content
    except Exception:
        pass
    return None


def add_cost_entry(item_id: str, cost_entry: dict) -> Optional[dict]:
    """
    Add a cost entry to an item and update total.

    Args:
        item_id: The history item ID
        cost_entry: Cost entry dict with operation, provider, model, cost_usd, timestamp

    Returns:
        Updated item or None if not found
    """
    history = load_history()

    for i, item in enumerate(history):
        if item.get('id') == item_id:
            # Ensure costs structure exists
            if 'costs' not in item:
                item['costs'] = {'entries': [], 'total_cost': 0.0}

            # Add entry
            item['costs']['entries'].append(cost_entry)
            item['costs']['total_cost'] += cost_entry.get('cost_usd', 0)
            item['costs']['total_cost'] = round(item['costs']['total_cost'], 6)

            history[i] = item
            save_history(history)
            return item

    return None


def add_qa_entry(item_id: str, question: str, answer: str, cost_info: dict = None, references: dict = None) -> Optional[dict]:
    """
    Add a Q&A exchange to an item's history.

    Args:
        item_id: The history item ID
        question: The question asked
        answer: The answer received
        cost_info: Optional cost info for this Q&A
        references: Optional references/citations

    Returns:
        Updated item or None if not found
    """
    history = load_history()

    for i, item in enumerate(history):
        if item.get('id') == item_id:
            # Ensure qa_history exists
            if 'qa_history' not in item:
                item['qa_history'] = []

            qa_entry = {
                'question': question,
                'answer': answer,
                'timestamp': datetime.now().isoformat()
            }

            if cost_info:
                qa_entry['cost_info'] = cost_info
            if references:
                qa_entry['references'] = references

            item['qa_history'].append(qa_entry)

            # Also add cost to costs if provided
            if cost_info:
                if 'costs' not in item:
                    item['costs'] = {'entries': [], 'total_cost': 0.0}
                cost_entry = {**cost_info, 'operation': 'qa'}
                item['costs']['entries'].append(cost_entry)
                item['costs']['total_cost'] += cost_entry.get('cost_usd', 0)
                item['costs']['total_cost'] = round(item['costs']['total_cost'], 6)

            history[i] = item
            save_history(history)
            return item

    return None


def save_resources(item_id: str, resources: dict, cost_info: dict = None) -> Optional[dict]:
    """
    Save resources/citations to an item.

    Args:
        item_id: The history item ID
        resources: Resources dict with answer, citations, etc.
        cost_info: Optional cost info for this request

    Returns:
        Updated item or None if not found
    """
    history = load_history()

    for i, item in enumerate(history):
        if item.get('id') == item_id:
            item['resources'] = resources

            # Add cost if provided
            if cost_info:
                if 'costs' not in item:
                    item['costs'] = {'entries': [], 'total_cost': 0.0}
                cost_entry = {**cost_info, 'operation': 'resources'}
                item['costs']['entries'].append(cost_entry)
                item['costs']['total_cost'] += cost_entry.get('cost_usd', 0)
                item['costs']['total_cost'] = round(item['costs']['total_cost'], 6)

            history[i] = item
            save_history(history)
            return item

    return None


def get_item_costs(item_id: str) -> dict:
    """Get all costs for an item."""
    item = get_history_item(item_id)
    if not item:
        return {'entries': [], 'total_cost': 0.0}
    return item.get('costs', {'entries': [], 'total_cost': 0.0})


def _migrate_analysis_format(item: dict) -> dict:
    """
    Migrate old single-analysis format to per-provider format.
    Called when loading items to ensure backwards compatibility.
    """
    analysis = item.get('analysis')
    if analysis is None:
        return item

    # Check if already in new format (dict with 'analyses' key)
    if isinstance(analysis, dict) and 'analyses' in analysis:
        return item

    # Old format: analysis is either a string or a dict with content
    # Convert to new format
    timestamp = item.get('analyzed_at', item.get('created_at', datetime.now().isoformat()))

    if isinstance(analysis, str):
        # Very old format: just a string
        new_analysis = {
            'analyses': {
                'unknown': {
                    'content': {'analysis': analysis},
                    'provider': 'unknown',
                    'timestamp': timestamp
                }
            },
            'primary_provider': 'unknown'
        }
    elif isinstance(analysis, dict):
        # Current format: dict with analysis, topics, provider, etc.
        provider = analysis.get('provider', 'unknown')
        new_analysis = {
            'analyses': {
                provider: {
                    'content': analysis,
                    'provider': provider,
                    'timestamp': timestamp,
                    'cost_info': analysis.get('cost_info')
                }
            },
            'primary_provider': provider
        }
    else:
        return item

    item['analysis'] = new_analysis
    return item


def save_provider_analysis(
    item_id: str,
    provider: str,
    analysis_content: dict,
    cost_info: dict = None,
    set_as_primary: bool = True
) -> Optional[dict]:
    """
    Save an analysis from a specific provider.

    Args:
        item_id: The history item ID
        provider: Provider name (openai, anthropic, google, etc.)
        analysis_content: The analysis result dict
        cost_info: Optional cost information
        set_as_primary: Whether to set this as the primary/default analysis

    Returns:
        Updated item or None if not found
    """
    history = load_history()

    for i, item in enumerate(history):
        if item.get('id') == item_id:
            # Ensure analysis structure exists
            if 'analysis' not in item or item['analysis'] is None:
                item['analysis'] = {'analyses': {}, 'primary_provider': None}

            # Migrate old format if needed
            item = _migrate_analysis_format(item)

            # Add this provider's analysis
            item['analysis']['analyses'][provider] = {
                'content': analysis_content,
                'provider': provider,
                'timestamp': datetime.now().isoformat(),
                'cost_info': cost_info
            }

            # Set as primary if requested or if it's the first analysis
            if set_as_primary or not item['analysis'].get('primary_provider'):
                item['analysis']['primary_provider'] = provider

            history[i] = item
            save_history(history)
            return item

    return None


def get_analysis_for_provider(item_id: str, provider: str) -> Optional[dict]:
    """
    Get analysis from a specific provider.

    Args:
        item_id: The history item ID
        provider: Provider name

    Returns:
        Provider's analysis dict or None
    """
    item = get_history_item(item_id)
    if not item:
        return None

    item = _migrate_analysis_format(item)
    analysis = item.get('analysis')
    if not analysis or 'analyses' not in analysis:
        return None

    return analysis['analyses'].get(provider)


def get_all_analyses(item_id: str) -> dict:
    """
    Get all provider analyses for an item.

    Args:
        item_id: The history item ID

    Returns:
        Dict with 'analyses' (by provider) and 'primary_provider'
    """
    item = get_history_item(item_id)
    if not item:
        return {'analyses': {}, 'primary_provider': None}

    item = _migrate_analysis_format(item)
    analysis = item.get('analysis')
    if not analysis or 'analyses' not in analysis:
        return {'analyses': {}, 'primary_provider': None}

    return {
        'analyses': analysis.get('analyses', {}),
        'primary_provider': analysis.get('primary_provider')
    }


def set_primary_analysis_provider(item_id: str, provider: str) -> Optional[dict]:
    """
    Set which provider's analysis should be shown by default.

    Args:
        item_id: The history item ID
        provider: Provider name to set as primary

    Returns:
        Updated item or None if not found
    """
    history = load_history()

    for i, item in enumerate(history):
        if item.get('id') == item_id:
            item = _migrate_analysis_format(item)
            analysis = item.get('analysis')
            if not analysis or 'analyses' not in analysis:
                return None

            # Check that this provider has an analysis
            if provider not in analysis['analyses']:
                return None

            item['analysis']['primary_provider'] = provider
            history[i] = item
            save_history(history)
            return item

    return None


def get_primary_analysis(item_id: str) -> Optional[dict]:
    """
    Get the primary (default) analysis for an item.

    Args:
        item_id: The history item ID

    Returns:
        Primary analysis content or None
    """
    item = get_history_item(item_id)
    if not item:
        return None

    item = _migrate_analysis_format(item)
    analysis = item.get('analysis')
    if not analysis or 'analyses' not in analysis:
        return None

    primary = analysis.get('primary_provider')
    if primary and primary in analysis['analyses']:
        provider_analysis = analysis['analyses'][primary]
        return provider_analysis.get('content')

    # Fallback to first analysis if no primary set
    if analysis['analyses']:
        first_provider = next(iter(analysis['analyses']))
        return analysis['analyses'][first_provider].get('content')

    return None


def get_full_item(item_id: str) -> Optional[dict]:
    """
    Get complete item data including transcription text, all analysis, Q&A history, resources, and costs.

    Args:
        item_id: The history item ID

    Returns:
        Full item dict or None if not found
    """
    from .language import detect_language_from_text

    item = get_history_item(item_id)
    if not item:
        return None

    # Migrate analysis format if needed
    item = _migrate_analysis_format(item)

    # Get transcription text
    text = get_transcription_text(item.get('file_path', ''))

    # Detect language if not already set
    language = item.get('language')
    text_direction = item.get('text_direction')

    if text and not text_direction:
        detected_lang, detected_dir = detect_language_from_text(text)
        if not language:
            language = detected_lang
        text_direction = detected_dir

        # Update the item in history if we detected language
        if detected_lang or detected_dir:
            update_history_item(item_id, language=language, text_direction=text_direction)

    # Get primary analysis content for backwards compatibility
    analysis_data = item.get('analysis')
    primary_analysis = None
    if analysis_data and 'analyses' in analysis_data:
        primary_provider = analysis_data.get('primary_provider')
        if primary_provider and primary_provider in analysis_data['analyses']:
            primary_analysis = analysis_data['analyses'][primary_provider].get('content')
        elif analysis_data['analyses']:
            # Fallback to first
            first_provider = next(iter(analysis_data['analyses']))
            primary_analysis = analysis_data['analyses'][first_provider].get('content')

    # Ensure all fields exist with defaults
    return {
        **item,
        'transcription': text,
        'language': language,
        'text_direction': text_direction or 'ltr',
        'qa_history': item.get('qa_history', []),
        'resources': item.get('resources'),
        'costs': item.get('costs', {'entries': [], 'total_cost': 0.0}),
        'analysis': primary_analysis,  # Primary analysis content for backwards compat
        'all_analyses': analysis_data,  # Full analysis collection
        'translations': item.get('translations', {})
    }


# =============================================================================
# v3 Project Migration Functions
# =============================================================================

def migrate_history_to_projects() -> dict:
    """
    Migrate all history items to the v3 project format.

    Returns:
        Dict with migration statistics
    """
    from .project import (
        load_projects, save_projects, migrate_history_item_to_project
    )

    history = load_history()
    existing_projects = load_projects()
    existing_ids = {p.get('video_id') for p in existing_projects}

    migrated = 0
    skipped = 0
    errors = 0

    for item in history:
        item_id = item.get('id')

        # Skip if already migrated
        if item_id in existing_ids:
            skipped += 1
            continue

        try:
            # Get transcription text
            text = get_transcription_text(item.get('file_path', ''))

            # Migrate to project format
            project = migrate_history_item_to_project(item, text or '')

            # Add to projects list
            existing_projects.insert(0, project.to_dict())
            existing_ids.add(item_id)
            migrated += 1
        except Exception as e:
            errors += 1

    # Save migrated projects
    if migrated > 0:
        save_projects(existing_projects)

    return {
        'total_history_items': len(history),
        'migrated': migrated,
        'skipped': skipped,
        'errors': errors
    }


def get_or_create_project(item_id: str) -> Optional[dict]:
    """
    Get a project for a history item, creating it if it doesn't exist.

    This provides backward compatibility - existing code can still use
    history items, but they'll be auto-converted to projects on access.

    Args:
        item_id: The history item / project ID

    Returns:
        Project dict or None if not found
    """
    from .project import get_project, save_project, migrate_history_item_to_project

    # First check if project already exists
    project = get_project(item_id)
    if project:
        return project.to_dict()

    # Get the history item
    item = get_history_item(item_id)
    if not item:
        return None

    # Migrate it to project format
    text = get_transcription_text(item.get('file_path', ''))
    project = migrate_history_item_to_project(item, text or '')

    # Save and return
    save_project(project)
    return project.to_dict()


def sync_history_item_to_project(item_id: str) -> Optional[dict]:
    """
    Sync changes from history item to its corresponding project.

    Call this after updating a history item to keep the project in sync.

    Args:
        item_id: The history item ID

    Returns:
        Updated project dict or None
    """
    from .project import get_project, save_project

    item = get_history_item(item_id)
    if not item:
        return None

    project = get_project(item_id)
    if not project:
        # Create new project
        return get_or_create_project(item_id)

    # Sync relevant fields from history to project
    text = get_transcription_text(item.get('file_path', ''))

    project.title = item.get('title', project.title)
    project.url = item.get('url', project.url)
    project.thumbnail = item.get('thumbnail', project.thumbnail)
    project.file_path = item.get('file_path', project.file_path)
    project.language = item.get('language', project.language)
    project.text_direction = item.get('text_direction', project.text_direction)
    project.transcription = text or project.transcription
    project.qa_history = item.get('qa_history', project.qa_history)
    project.resources = item.get('resources', project.resources)
    project.costs = item.get('costs', project.costs)

    # Sync analysis
    analysis_data = item.get('analysis')
    if analysis_data and isinstance(analysis_data, dict) and 'analyses' in analysis_data:
        project.analyses = analysis_data.get('analyses', {})
        project.primary_analysis_provider = analysis_data.get('primary_provider')

    save_project(project)
    return project.to_dict()
